#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import re
from collections import namedtuple
from os import path


def VERSION_PATTERNS_REGEX():
    return [re.compile(x) for x in [
        # Eclipse features
        r'v\d+\.feature\_(\d+\.){1,3}\d+',

        # Common version patterns
        r'(M?(v\d+(\-|\_))?\d+\.){1,3}\d+[A-Za-z0-9]*((\.|\-|_|~)'
            r'(b|B|rc|r|v|RC|alpha|beta|BETA|M|m|pre|vm|G)?\d+((\-|\.)\d+)?)?'
            r'((\.|\-)(((alpha|dev|beta|rc|FINAL|final|pre)(\-|\_)\d+[A-Za-z]?'
            r'(\-RELEASE)?)|alpha|dev(\.\d+\.\d+)?'
            r'|beta|BETA|final|FINAL|release|fixed|(cr\d(\_\d*)?)))?',
        #
        r'[A-Za-z]?(\d+\_){1,3}\d+\_?[A-Za-z]{0,2}\d+',
        #
        r'(b|rc|r|v|RC|alpha|beta|BETA|M|m|pre|revision-)\d+(\-\d+)?',
        #
        r'current|previous|latest|alpha|beta',
        #
        r'\d{4}-\d{2}-\d{2}',
        #
        r'(\d(\-|\_)){1,2}\d',
        #
        r'\d{5,14}',
    ]]


def hint(path):
    """
    Return a version found in a ``path`` or None. Prefix the version with 'v ' if
    the version does not start with v.
    """
    for pattern in VERSION_PATTERNS_REGEX():
        segments = path.split('/')
        # skip the first path segment unless there's only one segment
        first_segment = 1 if len(segments) > 1 else 0
        interesting_segments = segments[first_segment:]
        # we iterate backwards from the end of the paths segments list
        for segment in interesting_segments[::-1]:
            version = re.search(pattern, segment)
            if version:
                v = version.group(0)
                # prefix with v space
                if not v.lower().startswith('v'):
                    v = f'v {v}'
                return v


def is_dot_num(s):
    """
    Return True if a version string `s` is semver-like and composed only of dots
    and numbers.
    """
    return s.strip(".0123456789") == "" and not s.startswith(".") and not s.endswith(".")


common_version_suffixes = (
    "final",
    "release",
    "snapshot",
    "jre",
    "android",
    "pre",
    "alpha",
    "beta",
    "rc",
)
common_dash_version_suffixes = tuple(f"-{s}" for s in common_version_suffixes)


def is_moslty_num(s):
    """
    Return True if a version string `s` is primarily composed only of dots and
    numbers, with a minority of letters.

    >>> is_moslty_num("v11r2")
    True
    """
    dot_segments = s.split(".")
    len_alpha = 0
    len_digit = 0
    first_seg = dot_segments[0].lstrip("vV")
    starts_with_digit = first_seg.isdigit()
    for dot_seg in dot_segments:
        dot_seg = dot_seg.lstrip("vV")
        for seg in re.split("([0-9]+|[a-zA-Z]+)", dot_seg):
            if seg.isdigit():
                len_digit += len(seg)
            elif seg.isalpha() and seg.lower() not in common_version_suffixes:
                len_alpha += len(seg)

    if not len_alpha and not len_digit:
        return False

    if not len_alpha:
        return True

    # we want twice more digits than alphas
    if (2 * len_alpha) < len_digit:
        return True

    if starts_with_digit and len_alpha < len_digit:
        return True

    return False


NameVersion = namedtuple('NameVersion', 'name, version')


def get_jar_nv(filename):
    """
    Return a NameVersion tuple parsed from the JAR `filename` or None.

    For example::
        >>> get_jar_nv('org.eclipse.persistence.antlr_3.2.0.v201302191141.jar')
        NameVersion(name='org.eclipse.persistence.antlr', version='3.2.0.v201302191141')
        >>> get_jar_nv('org.eclipse.persistence.antlr.jar')
        NameVersion(name='org.eclipse.persistence.antlr', version=None)
        >>> get_jar_nv('org.eclipse.persistence.core_2.4.2.v20130514-5956486.jar')
        NameVersion(name='org.eclipse.persistence.core', version='2.4.2.v20130514-5956486')

        >>> get_jar_nv('com.io7m.jareas.checkstyle-0.2.2.jar')
        NameVersion(name='com.io7m.jareas.checkstyle', version='0.2.2')

        >>> get_jar_nv('ant-contrib-1.0b3.jar')
        NameVersion(name='ant-contrib', version='1.0b3')
        >>> get_jar_nv('xpp3-1.1.3.4.C.jar')
        NameVersion(name='xpp3', version='1.1.3.4.C')
        >>> get_jar_nv('ojdbc6_v11r2.jar')
        NameVersion(name='ojdbc6', version='v11r2')

        >>> get_jar_nv('amazon-sqs-java-messaging-lib-1.0.8.jar')
        NameVersion(name='amazon-sqs-java-messaging-lib', version='1.0.8')
        >>> get_jar_nv('annotations-4.1.1.4.jar')
        NameVersion(name='annotations', version='4.1.1.4')
        >>> get_jar_nv('aws-swf-build-tools-1.10.jar')
        NameVersion(name='aws-swf-build-tools', version='1.10')
        >>> get_jar_nv('c3p0-0.9.1.1.jar')
        NameVersion(name='c3p0', version='0.9.1.1')
        >>> get_jar_nv('javax.persistence_2.0.5.v201212031355.jar')
        NameVersion(name='javax.persistence', version='2.0.5.v201212031355')
        >>> get_jar_nv('proto-google-cloud-pubsub-v1-1.95.4.jar')
        NameVersion(name='proto-google-cloud-pubsub-v1', version='1.95.4')

        >>> get_jar_nv('xpp3-1.1.4c.jar')
        NameVersion(name='xpp3', version='1.1.4c')

        >>> get_jar_nv('listenablefuture-9999.0-empty-to-avoid-conflict-with-guava.jar')
        NameVersion(name='listenablefuture-9999.0-empty-to-avoid-conflict-with-guava', version=None)

        >>> get_jar_nv('aspectjweaver.jar')
        NameVersion(name='aspectjweaver', version=None)
        >>> get_jar_nv('flyway-client.jar')
        NameVersion(name='flyway-client', version=None)
        >>> get_jar_nv('jakarta.xml.bind-api.jar')
        NameVersion(name='jakarta.xml.bind-api', version=None)
        >>> get_jar_nv('javax.enterprise.concurrent.jar')
        NameVersion(name='javax.enterprise.concurrent', version=None)

        >>> get_jar_nv('netty-codec-http-4.1.53.Final.jar')
        NameVersion(name='netty-codec-http', version='4.1.53.Final')
        >>> get_jar_nv('spring-context-3.0.7.RELEASE.jar')
        NameVersion(name='spring-context', version='3.0.7.RELEASE')

        >>> get_jar_nv('guava-30.1-jre.jar')
        NameVersion(name='guava', version='30.1-jre')
        >>> get_jar_nv('guava-30.1.1-android.jar')
        NameVersion(name='guava', version='30.1.1-android')

        >>> get_jar_nv('guava-30.1.1-android.foo')

    """
    if not filename.endswith(".jar"):
        return

    basename, _extension = path.splitext(filename)

    # JAR name/version come in many flavors
    # amazon-sqs-java-messaging-lib-1.0.8.jar  is a plain name-ver
    if "_" in basename:
        # org.eclipse.persistence.antlr_3.2.0.v201302191141.jar
        name, _, version = basename.rpartition("_")
        if (
            is_dot_num(version)
            or is_moslty_num(version)
            or version.lower().endswith(common_version_suffixes)
        ):
            return NameVersion(name, version)

    if "-" in basename:
        # amazon-sqs-java-messaging-lib-1.0.8.jar
        dashname = basename
        suffix = ""
        for cs in common_dash_version_suffixes:
            if dashname.endswith(cs):
                dashname, _, suff = dashname.rpartition("-")
                suffix = f"-{suff}"
                break

        name, _, version = dashname.rpartition("-")
        if (
            is_dot_num(version)
            or is_moslty_num(version)
            or version.lower().endswith(common_version_suffixes)
        ):
            return NameVersion(name, f"{version}{suffix}")

    # no dash, no underscore means no version: org.eclipse.persistence.antlr.jar
    return NameVersion(basename, None)


def get_nupkg_nv(filename):
    """
    Return a NameVersion tuple parsed from the .nupkg NuGet archive `filename`.

    For example (taken from https://stackoverflow.com/questions/51662737/regex-to-parse-package-name-and-version-number-from-nuget-package-filenames/51662926):
        >>> get_nupkg_nv('knockoutjs.3.4.2.nupkg')
        NameVersion(name='knockoutjs', version='3.4.2')
        >>> get_nupkg_nv('log4net.2.0.8.nupkg')
        NameVersion(name='log4net', version='2.0.8')

        >>> get_nupkg_nv('runtime.tizen.4.0.0-armel.microsoft.netcore.jit.2.0.0.nupkg')
        NameVersion(name='runtime.tizen.4.0.0-armel.microsoft.netcore.jit', version='2.0.0')
        >>> get_nupkg_nv('nuget.core.2.7.0-alpha.nupkg')
        NameVersion(name='nuget.core', version='2.7.0-alpha')

        >>> get_nupkg_nv('microsoft.identitymodel.6.1.7600.16394.nupkg')
        NameVersion(name='microsoft.identitymodel', version='6.1.7600.16394')

        >>> get_nupkg_nv('guava.30.1.1.foo')
    """
    if not filename.endswith(".nupkg"):
        return

    basename, _extension = path.splitext(filename)

    # Either the last 3 or 4 segments are all digits in which case this is the
    # version. Otherwise we consider as version anything after the first all
    # digit segment starting from left.

    dot_segments = basename.split(".")
    len_dot_segments = len(dot_segments)
    if len_dot_segments > 4 and all(s.isdigit() for s in dot_segments[-4:]):
        names = dot_segments[:-4]
        versions = dot_segments[-4:]
        return NameVersion(".".join(names), ".".join(versions))

    if len_dot_segments > 3 and all(s.isdigit() for s in dot_segments[-3:]):
        names = dot_segments[:-3]
        versions = dot_segments[-3:]
        return NameVersion(".".join(names), ".".join(versions))

    names = []
    versions = []
    in_version = False
    for seg in dot_segments:
        if in_version:
            versions.append(seg)
            continue

        if not seg.isdigit():
            names.append(seg)
        else:
            versions.append(seg)
            in_version = True

    return NameVersion(".".join(names), ".".join(versions))



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import unittest

from commoncode import version


class TestVersionHint(unittest.TestCase):

    def test_version_hint(self):
        data = {
            '/xmlgraphics/fop/source/fop-1.0-src.zip': '1.0',
            '/xml/xindice/xml-xindice-1.2m1-src.zip': '1.2m1',
            '/xmlgraphics/fop/binaries/fop-0.94-bin-jdk1.3.tar.gz': '0.94',
            '/xmlgraphics/batik/batik-src-1.7beta1.zip': '1.7beta1',
            '/xmlgraphics/batik/batik-1.7-jre13.zip': '1.7',
            '/xmlbeans/source/xmlbeans-2.3.0-src.tgz': '2.3.0',
            '/xml/xindice/source/xml-xindice-1.2m1-src.tar.gz': '1.2m1',
            '/xml/xerces-p/binaries/XML-Xerces-2.3.0-4-win32.zip': '2.3.0-4',
            '/xml/xerces-p/source/XML-Xerces-2.3.0-3.tar.gz': '2.3.0-3',
            '/xml/xalan-j/source/xalan-j_2_7_0-src-2jars.tar.gz': '2_7_0',
            '/xml/security/java-library/xml-security-src-1_0_5D2.zip': '1_0_5D2',
            '/xml/commons/binaries/xml-commons-external-1.4.01-bin.zip': '1.4.01',
            '/xml/commons/xml-commons-1.0.b2.zip': '1.0.b2',
            '/xml/cocoon/3.0/cocoon-all-3.0.0-alpha-1-dist.tar.gz': '3.0.0-alpha-1',
            '/xerces/j/source/Xerces-J-tools.2.10.0-xml-schema-1.1-beta.tar.gz': '2.10.0',
            '/xerces/c/3/binaries/xerces-c-3.1.1-x86_64-solaris-cc-5.10.tar.gz': '3.1.1',
            '/xerces/c/3/binaries/xerces-c-3.1.1-x86_64-windows-vc-8.0.zip': '3.1.1',
            '/xerces/c/2/binaries/xerces-c_2_8_0-x86-windows-vc_7_1.zip': '2_8_0',
            '/ws/woden/1.0M8/apache-woden-src-1.0M8.tar.gz': '1.0M8',
            '/ws/scout/0_7rc1/source/scout-0.7rc1-src.zip': '0.7rc1',
            '/ws/juddi/3_0/juddi-portal-bundle-3.0.0.rc1.zip': '3.0.0.rc1',
            '/ws/juddi/3_0/juddi-portal-bundle-3.0.0.beta.zip': '3.0.0.beta',
            '/ws/juddi/2_0RC7/juddi-tomcat-2.0rc7.zip': '2.0rc7',
            '/ws/axis2/tools/1_4_1/axis2-wsdl2code-maven-plugin-1.4.1.jar': '1.4.1',
            '/ws/axis/1_4/axis-src-1_4.zip': '1_4',
            '/ws/axis-c/source/win32/axis-c-1.6b-Win32-trace-src.zip': '1.6b',
            '/tuscany/java/sca/2.0-M5/apache-tuscany-sca-all-2.0-M5-src.tar.gz': '2.0-M5',
            '/turbine/turbine-2.3.3-rc1/source/turbine-2.3.3-RC1-src.zip': '2.3.3-RC1',
            '/tomcat/tomcat-connectors/jk/binaries/win64/jk-1.2.30/ia64/symbols-1.2.30.zip': '1.2.30',
            '/tomcat/tomcat-7/v7.0.0-beta/bin/apache-tomcat-7.0.0-windows-i64.zip': '7.0.0',
            '/tomcat/tomcat-4/v4.1.40/bin/apache-tomcat-4.1.40-LE-jdk14.exe': '4.1.40',
            '/tapestry/tapestry-src-5.1.0.5.tar.gz': '5.1.0.5',
            '/spamassassin/source/Mail-SpamAssassin-rules-3.3.0.r901671.tgz': '3.3.0.r901671',
            '/spamassassin/Mail-SpamAssassin-rules-3.3.1.r923257.tgz': '3.3.1.r923257',
            '/shindig/1.1-BETA5-incubating/shindig-1.1-BETA5-incubating-source.zip': '1.1-BETA5',
            '/servicemix/nmr/1.0.0-m3/apache-servicemix-nmr-1.0.0-m3-src.tar.gz': '1.0.0-m3',
            '/qpid/0.6/qpid-dotnet-0-10-0.6.zip': '0.6',
            '/openjpa/2.0.0-beta/apache-openjpa-2.0.0-beta-binary.zip': '2.0.0-beta',
            '/myfaces/source/portlet-bridge-2.0.0-alpha-2-src-all.tar.gz': '2.0.0-alpha-2',
            '/myfaces/source/myfaces-extval20-2.0.3-src.tar.gz': '2.0.3',
            '/harmony/milestones/6.0/debian/amd64/harmony-6.0-classlib_0.0r946981-1_amd64.deb': '6.0',
            '/geronimo/eclipse/updates/plugins/org.apache.geronimo.st.v21.ui_2.1.1.jar': '2.1.1',
            '/directory/studio/update/1.x/plugins/org.apache.directory.studio.aciitemeditor_1.5.2.v20091211.jar': '1.5.2.v20091211',
            '/db/torque/torque-3.3/source/torque-gen-3.3-RC3-src.zip': '3.3-RC3',
            '/cayenne/cayenne-3.0B1.tar.gz': '3.0B1',
            '/cayenne/cayenne-3.0M4-macosx.dmg': '3.0M4',
            '/xmlgraphics/batik/batik-docs-current.zip': 'current',
            '/xmlgraphics/batik/batik-docs-previous.zip': 'previous',
            '/poi/dev/bin/poi-bin-3.7-beta1-20100620.zip': '3.7-beta1-20100620',
            '/excalibur/avalon-logkit/source/excalibur-logkit-2.0.dev-0-src.zip': '2.0.dev-0',
            '/db/derby/db-derby-10.4.2.0/derby_core_plugin_10.4.2.zip': '10.4.2',
            '/httpd/modpython/win/2.7.1/mp152dll.zip': '2.7.1',
            '/perl/mod_perl-1.31/apaci/mod_perl.config.sh': '1.31',
            '/xml/xerces-j/old_xerces2/Xerces-J-bin.2.0.0.alpha.zip': '2.0.0.alpha',
            '/xml/xerces-p/archives/XML-Xerces-1.7.0_0.tar.gz': '1.7.0_0',
            '/httpd/docs/tools-2004-05-04.zip': '2004-05-04',
            '/ws/axis2/c/M0_5/axis2c-src-M0.5.tar.gz': 'M0.5',
            '/jakarta/poi/dev/src/jakarta-poi-1.8.0-dev-src.zip': '1.8.0-dev',
            '/tapestry/tapestry-4.0-beta-8.zip': '4.0-beta-8',
            '/openejb/3.0-beta-1/openejb-3.0-beta-1.zip': '3.0-beta-1',
            '/tapestry/tapestry-4.0-rc-1.zip': '4.0-rc-1',
            '/jakarta/tapestry/source/3.0-rc-3/Tapestry-3.0-rc-3-src.zip': '3.0-rc-3',
            '/jakarta/lucene/binaries/lucene-1.3-final.tar.gz': '1.3-final',
            '/jakarta/tapestry/binaries/3.0-beta-1a/Tapestry-3.0-beta-1a-bin.zip': '3.0-beta-1a',
            '/poi/release/bin/poi-bin-3.0-FINAL-20070503.tar.gz': '3.0-FINAL-20070503',
            '/harmony/milestones/M4/apache-harmony-hdk-r603534-linux-x86-32-libstdc++v6-snapshot.tar.gz': 'r603534',
            '/ant/antidote/antidote-20050330.tar.bz2': '20050330',
            '/apr/not-released/apr_20020725223645.tar.gz': '20020725223645',
            '/ibatis/source/ibatis.net/src-revision-709676.zip': 'revision-709676',
            '/ws/axis-c/source/win32/axis-c-src-1-2-win32.zip': '1-2',
            '/jakarta/slide/most-recent-2.0rc1-binaries/jakarta-slide 2.0rc1 jakarta-tomcat-4.1.30.zip': '2.0rc1',
            '/httpd/modpython/win/3.0.1/python2.2.1-apache2.0.43.zip': '2.2.1',
            '/ant/ivyde/updatesite/features/org.apache.ivy.feature_2.1.0.cr1_20090319213629.jar': '2.1.0.cr1_20090319213629',
            '/jakarta/poi/dev/bin/poi-2.0-pre1-20030517.jar': '2.0-pre1-20030517',
            '/jakarta/poi/release/bin/jakarta-poi-1.5.0-FINAL-bin.zip': '1.5.0-FINAL',
            '/jakarta/poi/release/bin/poi-bin-2.0-final-20040126.zip': '2.0-final-20040126',
            '/activemq/apache-activemq/5.0.0/apache-activemq-5.0.0-sources.jar': '5.0.0',
            '/turbine/turbine-2.2/source/jakarta-turbine-2.2-B1.tar.gz': '2.2-B1',
            '/ant/ivyde/updatesite/features/org.apache.ivy.feature_2.0.0.cr1.jar': '2.0.0.cr1',
            '/ant/ivyde/updatesite/features/org.apache.ivy.feature_2.0.0.final_20090108225011.jar': '2.0.0.final_20090108225011',
            '/ws/axis/1_2RC3/axis-src-1_2RC3.zip': '1_2RC3',
            '/commons/lang/old/v1.0-b1.1/commons-lang-1.0-b1.1.zip': '1.0-b1.1',
            '/commons/net/binaries/commons-net-1.2.0-release.tar.gz': '1.2.0-release',
            '/ant/ivyde/2.0.0.final/apache-ivyde-2.0.0.final-200907011148-RELEASE.tgz': '2.0.0.final-200907011148-RELEASE',
            '/geronimo/eclipse/updates/plugins/org.apache.geronimo.jetty.j2ee.server.v11_1.0.0.jar': 'v11_1.0.0',
            '/jakarta/cactus/binaries/jakarta-cactus-13-1.7.1-fixed.zip': '1.7.1-fixed',
            '/jakarta/jakarta-turbine-maven/maven/jars/maven-1.0-b5-dev.20020731.085427.jar': '1.0-b5-dev.20020731.085427',
            '/xml/xalan-j/source/xalan-j_2_5_D1-src.tar.gz': '2_5_D1',
            '/ws/woden/IBuilds/I20051002_1145/woden-I20051002_1145.tar.bz2': 'I20051002_1145',
            '/commons/beanutils/source/commons-beanutils-1.8.0-BETA-src.tar.gz': '1.8.0-BETA',
            '/cocoon/BINARIES/cocoon-2.0.3-vm14-bin.tar.gz': '2.0.3-vm14',
            '/felix/xliff_filters_v1_2_7_unix.jar': 'v1_2_7',
            '/excalibur/releases/200702/excalibur-javadoc-r508111-15022007.tar.gz': 'r508111-15022007',
            '/geronimo/eclipse/updates/features/org.apache.geronimo.v20.feature_2.0.0.jar': 'v20.feature_2.0.0',
            '/geronimo/2.1.6/axis2-jaxws-1.3-G20090406.jar': '1.3-G20090406',
            '/cassandra/debian/pool/main/c/cassandra/cassandra_0.4.0~beta1-1.diff.gz': '0.4.0~beta1-1',
            '/ha-api-3.1.6.jar': '3.1.6',
            'ha-api-3.1.6.jar': '3.1.6'
            }

        # FIXME: generate a test function for each case
        for path in data:
            expected = data[path]
            if not expected.lower().startswith('v'):
                expected = 'v ' + expected
            assert version.hint(path) == expected



#
# Copyright (c) nexB Inc. and others. All rights reserved.
# SPDX-License-Identifier: Apache-2.0 AND Python-2.0
# See http://www.apache.org/licenses/LICENSE-2.0 for the license text.
# See https://github.com/nexB/commoncode for support or download.
# See https://aboutcode.org for more information about nexB OSS projects.
#

import errno
import os
import ntpath
import posixpath
import shutil
import stat
import sys
import tempfile

from os import fsdecode

try:
    from scancode_config import scancode_temp_dir as _base_temp_dir
except ImportError:
    _base_temp_dir = None

from commoncode import filetype
from commoncode.filetype import is_rwx

# this exception is not available on posix
try:
    WindowsError  # NOQA
except NameError:

    class WindowsError(Exception):
        pass

import logging
logger = logging.getLogger(__name__)

TRACE = False


def logger_debug(*args):
    pass


if TRACE:
    logging.basicConfig(stream=sys.stdout)
    logger.setLevel(logging.DEBUG)

    def logger_debug(*args):
        return logger.debug(' '.join(isinstance(a, str) and a or repr(a) for a in args))

"""
File, paths and directory utility functions.
"""

#
# DIRECTORIES
#


def create_dir(location):
    """
    Create directory and all sub-directories recursively at location ensuring these
    are readable and writeable.
    Raise Exceptions if it fails to create the directory.
    """

    if os.path.exists(location):
        if not os.path.isdir(location):
            err = ('Cannot create directory: existing file '
                   'in the way ''%(location)s.')
            raise OSError(err % locals())
    else:
        # may fail on win if the path is too long
        # FIXME: consider using UNC ?\\ paths

        try:
            os.makedirs(location)
            chmod(location, RW, recurse=False)

        # avoid multi-process TOCTOU conditions when creating dirs
        # the directory may have been created since the exist check
        except WindowsError as e:
            # [Error 183] Cannot create a file when that file already exists
            if e and e.winerror == 183:
                if not os.path.isdir(location):
                    raise
            else:
                raise
        except (IOError, OSError) as o:
            if o.errno == errno.EEXIST:
                if not os.path.isdir(location):
                    raise
            else:
                raise


def get_temp_dir(base_dir=_base_temp_dir, prefix=''):
    """
    Return the path to a new existing unique temporary directory, created under
    the `base_dir` base directory using the `prefix` prefix.
    If `base_dir` is not provided, use the 'SCANCODE_TMP' env var or the system
    temp directory.

    WARNING: do not change this code without changing scancode_config.py too
    """

    has_base = bool(base_dir)
    if not has_base:
        base_dir = os.getenv('SCANCODE_TMP')
        if not base_dir:
            base_dir = tempfile.gettempdir()

    if not os.path.exists(base_dir):
        create_dir(base_dir)

    if not has_base:
        prefix = 'scancode-tk-'

    return tempfile.mkdtemp(prefix=prefix, dir=base_dir)

#
# PATHS AND NAMES MANIPULATIONS
#

# TODO: move these functions to paths.py


def prepare_path(pth):
    """
    Return the `pth` path string either as encoded bytes if on Linux and using
    Python 2 or as a unicode/text otherwise.
    """
    if not isinstance(pth, str):
        return fsdecode(pth)
    return pth


def is_posixpath(location):
    """
    Return True if the `location` path is likely a POSIX-like path using POSIX path
    separators (slash or "/")or has no path separator.

    Return False if the `location` path is likely a Windows-like path using backslash
    as path separators (e.g. "\").
    """
    has_slashes = '/' in location
    has_backslashes = '\\' in location
    # windows paths with drive
    if location:
        drive, _ = ntpath.splitdrive(location)
        if drive:
            return False

    # a path is always POSIX unless it contains ONLY backslahes
    # which is a rough approximation (it could still be posix)
    is_posix = True
    if has_backslashes and not has_slashes:
        is_posix = False
    return is_posix


def as_posixpath(location):
    """
    Return a POSIX-like path using POSIX path separators (slash or "/") for a
    `location` path. This converts Windows paths to look like POSIX paths: Python
    accepts gracefully POSIX paths on Windows.
    """
    location = prepare_path(location)
    return location.replace('\\', '/')


def as_winpath(location):
    """
    Return a Windows-like path using Windows path separators (backslash or "\") for a
    `location` path.
    """
    location = prepare_path(location)
    return location.replace('/', '\\')


def split_parent_resource(path, force_posix=False):
    """
    Return a tuple of (parent directory path, resource name).
    """
    use_posix = force_posix or is_posixpath(path)
    splitter = use_posix and posixpath or ntpath
    path_no_trailing_speps = path.rstrip('\\/')
    return splitter.split(path_no_trailing_speps)


def resource_name(path, force_posix=False):
    """
    Return the resource name (file name or directory name) from `path` which
    is the last path segment.
    """
    _left, right = split_parent_resource(path, force_posix)
    return right or ''


def file_name(path, force_posix=False):
    """
    Return the file name (or directory name) of a path.
    """
    return resource_name(path, force_posix)


def parent_directory(path, force_posix=False, with_trail=True):
    """
    Return the parent directory path of a file or directory `path`.
    The returned directory has a trailing path separator unless with_trail is False.
    """
    left, _right = split_parent_resource(path, force_posix)
    use_posix = force_posix or is_posixpath(path)
    sep = '/' if use_posix else '\\'
    trail = sep if with_trail and left != sep else ''
    return left + trail


def file_base_name(path, force_posix=False):
    """
    Return the file base name for a path. The base name is the base name of
    the file minus the extension. For a directory return an empty string.
    """
    return splitext(path, force_posix)[0]


def file_extension(path, force_posix=False):
    """
    Return the file extension for a path.
    """
    return splitext(path, force_posix)[1]


def splitext_name(file_name, is_file=True):
    """
    Return a tuple of Unicode strings (basename, extension) for a file name. The
    basename is the file name minus its extension. Return an empty extension
    string for a directory. Not the same as os.path.splitext_name.
    """

    if not file_name:
        return '', ''
    file_name = fsdecode(file_name)

    if not is_file:
        return file_name, ''

    if file_name.startswith('.') and '.' not in file_name[1:]:
        # .dot files base name is the full name and they do not have an extension
        return file_name, ''

    base_name, extension = posixpath.splitext(file_name)
    # handle composed extensions of tar.gz, bz, zx,etc
    if base_name.endswith('.tar'):
        base_name, extension2 = posixpath.splitext(base_name)
        extension = extension2 + extension
    return base_name, extension


# TODO: FIXME: this is badly broken!!!!
def splitext(path, force_posix=False):
    """
    Return a tuple of strings (basename, extension) for a path. The basename is
    the file name minus its extension. Return an empty extension string for a
    directory.
    """
    base_name = ''
    extension = ''
    if not path:
        return base_name, extension

    is_dir = path.endswith(('\\', '/',))
    path = as_posixpath(path).strip('/')
    name = resource_name(path, force_posix)
    if is_dir:
        # directories never have an extension
        base_name = name
        extension = ''
    elif name.startswith('.') and '.' not in name[1:]:
        # .dot files base name is the full name and they do not have an extension
        base_name = name
        extension = ''
    else:
        base_name, extension = posixpath.splitext(name)
        # handle composed extensions of tar.gz, tar.bz2, zx,etc
        if base_name.endswith('.tar'):
            base_name, extension2 = posixpath.splitext(base_name)
            extension = extension2 + extension
    return base_name, extension

#
# DIRECTORY AND FILES WALKING/ITERATION
#


def ignore_nothing(_):
    return False


def walk(location, ignored=None, follow_symlinks=False):
    """
    Walk location returning the same tuples as os.walk but with a different
    behavior:
     - always walk top-down, breadth-first.
     - ignore and do not follow symlinks unless `follow_symlinks` is True,
     - always ignore special files (FIFOs, etc.)
     - optionally ignore files and directories by invoking the `ignored`
       callable on files and directories returning True if it should be ignored.
     - location is a directory or a file: for a file, the file is returned.

    If `follow_symlinks` is True, then symlinks will not be ignored and be
    collected like regular files and directories
    """
    # TODO: consider using the new "scandir" module for some speed-up.

    is_ignored = ignored(location) if ignored else False
    if is_ignored:
        if TRACE:
            logger_debug('walk: ignored:', location, is_ignored)
        return

    if filetype.is_file(location, follow_symlinks=follow_symlinks) :
        yield parent_directory(location), [], [file_name(location)]

    elif filetype.is_dir(location, follow_symlinks=follow_symlinks):
        dirs = []
        files = []
        # TODO: consider using scandir
        for name in os.listdir(location):
            loc = os.path.join(location, name)
            if filetype.is_special(loc) or (ignored and ignored(loc)):
                if (follow_symlinks
                        and filetype.is_link(loc)
                        and not filetype.is_broken_link(location)):
                    pass
                else:
                    if TRACE:
                        ign = ignored and ignored(loc)
                        logger_debug('walk: ignored:', loc, ign)
                    continue
            # special files and symlinks are always ignored
            if filetype.is_dir(loc, follow_symlinks=follow_symlinks):
                dirs.append(name)
            elif filetype.is_file(loc, follow_symlinks=follow_symlinks):
                files.append(name)
        yield location, dirs, files

        for dr in dirs:
            for tripple in walk(os.path.join(location, dr), ignored, follow_symlinks=follow_symlinks):
                yield tripple


def resource_iter(location, ignored=ignore_nothing, with_dirs=True, follow_symlinks=False):
    """
    Return an iterable of paths at `location` recursively.

    :param location: a file or a directory.
    :param ignored: a callable accepting a location argument and returning True
                    if the location should be ignored.
    :return: an iterable of file and directory locations.
    """
    for top, dirs, files in walk(location, ignored, follow_symlinks=follow_symlinks):
        if with_dirs:
            for d in dirs:
                yield os.path.join(top, d)
        for f in files:
            yield os.path.join(top, f)
#
# COPY
#


def copytree(src, dst):
    """
    Copy recursively the `src` directory to the `dst` directory. If `dst` is an
    existing directory, files in `dst` may be overwritten during the copy.
    Preserve timestamps.
    Ignores:
     -`src` permissions: `dst` files are created with the default permissions.
     - all special files such as FIFO or character devices and symlinks.

    Raise an shutil.Error with a list of reasons.

    This function is similar to and derived from the Python shutil.copytree
    function. See fileutils.py.ABOUT for details.
    """
    if not filetype.is_readable(src):
        chmod(src, R, recurse=False)

    names = os.listdir(src)

    if not os.path.exists(dst):
        os.makedirs(dst)

    errors = []
    errors.extend(copytime(src, dst))

    for name in names:
        srcname = os.path.join(src, name)
        dstname = os.path.join(dst, name)

        # skip anything that is not a regular file, dir or link
        if not filetype.is_regular(srcname):
            continue

        if not filetype.is_readable(srcname):
            chmod(srcname, R, recurse=False)
        try:
            if os.path.isdir(srcname):
                copytree(srcname, dstname)
            elif filetype.is_file(srcname):
                copyfile(srcname, dstname)
        # catch the Error from the recursive copytree so that we can
        # continue with other files
        except shutil.Error as err:
            errors.extend(err.args[0])
        except EnvironmentError as why:
            errors.append((srcname, dstname, str(why)))

    if errors:
        raise shutil.Error(errors)


def copyfile(src, dst):
    """
    Copy src file to dst file preserving timestamps.
    Ignore permissions and special files.

    Similar to and derived from Python shutil module. See fileutils.py.ABOUT
    for details.
    """
    if not filetype.is_regular(src):
        return
    if not filetype.is_readable(src):
        chmod(src, R, recurse=False)
    if os.path.isdir(dst):
        dst = os.path.join(dst, os.path.basename(src))
    shutil.copyfile(src, dst)
    copytime(src, dst)


def copytime(src, dst):
    """
    Copy timestamps from `src` to `dst`.

    Similar to and derived from Python shutil module. See fileutils.py.ABOUT
    for details.
    """
    errors = []
    st = os.stat(src)
    if hasattr(os, 'utime'):
        try:
            os.utime(dst, (st.st_atime, st.st_mtime))
        except OSError as why:
            if WindowsError is not None and isinstance(why, WindowsError):
                # File access times cannot be copied on Windows
                pass
            else:
                errors.append((src, dst, str(why)))
    return errors

#
# PERMISSIONS
#


# modes: read, write, executable
R = stat.S_IRUSR
RW = stat.S_IRUSR | stat.S_IWUSR
RX = stat.S_IRUSR | stat.S_IXUSR
RWX = stat.S_IRUSR | stat.S_IWUSR | stat.S_IXUSR


# FIXME: This was an expensive operation that used to recurse of the parent directory
def chmod(location, flags, recurse=False):
    """
    Update permissions for `location` with with `flags`. `flags` is one of R,
    RW, RX or RWX with the same semantics as in the chmod command. Update is
    done recursively if `recurse`.
    """
    if not location or not os.path.exists(location):
        return
    location = os.path.abspath(location)

    new_flags = flags
    if filetype.is_dir(location):
        # POSIX dirs need to be executable to be readable,
        # and to be writable so we can change perms of files inside
        new_flags = RWX

    # FIXME: do we really need to change the parent directory perms?
    # FIXME: may just check them instead?
    parent = os.path.dirname(location)
    current_stat = stat.S_IMODE(os.stat(parent).st_mode)
    if not is_rwx(parent):
        os.chmod(parent, current_stat | RWX)

    if filetype.is_regular(location):
        current_stat = stat.S_IMODE(os.stat(location).st_mode)
        os.chmod(location, current_stat | new_flags)

    if recurse:
        chmod_tree(location, flags)


def chmod_tree(location, flags):
    """
    Update permissions recursively in a directory tree `location`.
    """
    if filetype.is_dir(location):
        for top, dirs, files in walk(location):
            for d in dirs:
                chmod(os.path.join(top, d), flags, recurse=False)
            for f in files:
                chmod(os.path.join(top, f), flags, recurse=False)

#
# DELETION
#


def _rm_handler(function, path, excinfo):  # NOQA
    """
    shutil.rmtree handler invoked on error when deleting a directory tree.
    This retries deleting once before giving up.
    """
    if TRACE:
        logger_debug('_rm_handler:', 'path:', path, 'excinfo:', excinfo)
    if function in (os.rmdir, os.listdir):
        try:
            chmod(path, RW, recurse=True)
            shutil.rmtree(path, True)
        except Exception:
            pass

        if os.path.exists(path):
            logger.warning('Failed to delete directory %s', path)

    elif function == os.remove:
        try:
            delete(path, _err_handler=None)
        except:
            pass

        if os.path.exists(path):
            logger.warning('Failed to delete file %s', path)


def delete(location, _err_handler=_rm_handler):
    """
    Delete a directory or file at `location` recursively. Similar to "rm -rf"
    in a shell or a combo of os.remove and shutil.rmtree.
    This function is design to never fails: instead it logs warnings if the
    file or directory was not deleted correctly.
    """
    if not location:
        return

    if os.path.exists(location) or filetype.is_broken_link(location):
        if filetype.is_dir(location):
            shutil.rmtree(location, False, _rm_handler)
        else:
            os.remove(location)
